/*
 * mylink.c
 *
 * Functions returning results from a remote mysql database
 *
 * liuguangzhao@users.sf.net
 *
 * contrib/mylink/mylink.c
 * Copyright (c) 2010-2020, kitsoft.com
 * ALL RIGHTS RESERVED;
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR OR DISTRIBUTORS BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHOR OR DISTRIBUTORS HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHOR AND DISTRIBUTORS HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 */

#include "postgres.h"

#include <limits.h>
#include <assert.h>

#include "libpq-fe.h"
#include "fmgr.h"
#include "funcapi.h"
#include "access/genam.h"
#include "access/heapam.h"
#include "access/tupdesc.h"
#include "catalog/indexing.h"
#include "catalog/namespace.h"
#include "catalog/pg_index.h"
#include "catalog/pg_type.h"
#include "executor/executor.h"
#include "executor/spi.h"
#include "foreign/foreign.h"
#include "lib/stringinfo.h"
#include "mb/pg_wchar.h"
#include "miscadmin.h"
#include "nodes/execnodes.h"
#include "nodes/nodes.h"
#include "nodes/pg_list.h"
#include "parser/parse_type.h"
#include "parser/scansup.h"
#include "utils/acl.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "utils/dynahash.h"
// #include "utils/fmgroids.h" #pgsql-9.1 only
#include "utils/palloc.h"
#include "utils/hsearch.h"
#include "utils/lsyscache.h"
#include "utils/memutils.h"
#include "utils/syscache.h"
#include "utils/tqual.h"

#include "mylink.h"

#include "mysql.h"
// fix LIST struct conflict with same name function in nodes/pg_list.h 
// list_delete,list_free,list_length
// we use myself modified version mysql.h and my_list.h
// if this has problem, delete current mysql.h and my_list.h
// then modify you my_list.h, change the three function names to someone else

PG_MODULE_MAGIC;

typedef struct RemoteConn
{
	MYSQL   conn;			/* Hold the remote connection */
	int	    openCursorCount;	/* The number of open cursors */
	bool	newXactForCursor;		/* Opened a transaction for a cursor */

    unsigned int last_errno;
    char    last_error[100];
} RemoteConn;

static RemoteConn *getConnectionByName(const char *conn_name);
static HTAB *createConnHash(void);
static void createNewConnection(const char *conn_name, RemoteConn *rconn);
static void deleteConnection(const char *conn_name);


/* Global */
// static RemoteConn *pconn = NULL;
static HTAB *hconn = NULL;
static const char *default_conn_name = "mylink_default_conn_name_db1b_f18b_e9d2_cc29";


/*
 *	Following is list that holds multiple remote connections.
 *	Calling convention of each dblink function changes to accept
 *	connection name as the first parameter. The connection list is
 *	much like ecpg e.g. a mapping between a name and a PGconn object.
 */

typedef struct remoteConnHashEnt
{
	char		name[NAMEDATALEN];
	RemoteConn *rconn;
} RemoteConnHashEnt;

/* initial number of connection hashes */
#define NUMCONN 16


// dummy function for link libpq default
static void mylink_dummy_func()
{
    PQstatus(NULL);
}

static char *mylink_last_error(const char *conn_name)
{
    RemoteConn *tconn;
    char *emsg;

    tconn = getConnectionByName(conn_name);
    if (tconn == NULL) {
        return NULL;
    }
  
    memset(tconn->last_error, 0, sizeof(tconn->last_error));
    snprintf(tconn->last_error, sizeof(tconn->last_error)-1, "EMYLINK: %s", mysql_error(&tconn->conn));
    tconn->last_errno = mysql_errno(&tconn->conn);

    emsg = tconn->last_error;

    return emsg;
}

/*
 * Create a persistent connection to another database
 */
PG_FUNCTION_INFO_V1(mylink_connect);
Datum
mylink_connect(PG_FUNCTION_ARGS)
{
    RemoteConn *tconn;
    int need_connect;
    MYSQL *conn;
    MYSQL *rconn;

    char *host;
    int port;
    char *user;
    char *passwd;
    char *dbname;
    char *conn_name;

    // check args validation
    if (PG_NARGS() == 4 || PG_NARGS() == 5 || PG_NARGS() == 6) {
    } else {
        PG_RETURN_TEXT_P(cstring_to_text("ERR"));
    }

    host = text_to_cstring(PG_GETARG_TEXT_PP(0));
    port = PG_GETARG_INT32(1);
    if (port <= 0 || port > 65535) port = 5432; 
    user = text_to_cstring(PG_GETARG_TEXT_PP(2));
    passwd = text_to_cstring(PG_GETARG_TEXT_PP(3));
    dbname = NULL;
    conn_name = default_conn_name;

    if (PG_NARGS() == 6) {
        dbname = text_to_cstring(PG_GETARG_TEXT_PP(4));
        conn_name = text_to_cstring(PG_GETARG_TEXT_PP(5));
    }
    if (PG_NARGS() == 5) {
        dbname = text_to_cstring(PG_GETARG_TEXT_PP(4));
    }

    if (host == NULL || user == NULL || strlen(host) == 0 || strlen(user) == 0) {
        ereport(WARNING, (errcode(-1), errmsg("EMYLINK: Invalid arguments.")));
        PG_RETURN_TEXT_P(cstring_to_text("ERR"));
    }

    need_connect = 0; 
    my_init();
    tconn = getConnectionByName(conn_name);
    if (tconn == NULL) {
        tconn = (RemoteConn *) MemoryContextAlloc(TopMemoryContext, sizeof(RemoteConn));
        conn = mysql_init(&(tconn->conn));
        if (conn == NULL) {
            ereport(ERROR,
                    (errcode(ERRCODE_SQLCLIENT_UNABLE_TO_ESTABLISH_SQLCONNECTION),
                     errmsg("could not establish connectioncccccccccccccc"),
                     errdetail("%s", "aaaaaaaaa")));
        } else {
        }
        need_connect = 1;
        createNewConnection(conn_name, tconn);
    }
  
    if (need_connect == 0) {
        if (mysql_ping(&(tconn->conn)) == 0) {
        } else {
            ereport(WARNING,
                    (errcode(ERRCODE_SQLCLIENT_UNABLE_TO_ESTABLISH_SQLCONNECTION),
                     errmsg("Connection to mysql is lost."),
                     errdetail("%s", "aaaaaaaaa")));
            need_connect = 1;
        }
    } 

    if (need_connect == 1) {

        // rconn = mysql_real_connect(&tconn->conn, "127.0.0.1", "root", "posta", "dzforum", 0, NULL, 0);
        rconn = mysql_real_connect(&tconn->conn, host, user, passwd, dbname, 0, NULL, 0);
        if (rconn == NULL) {
            ereport(WARNING, (errcode(-1), errmsg(mylink_last_error(conn_name))));
            PG_RETURN_TEXT_P(cstring_to_text("ERR"));
        } else {
        }
    }
  
    PG_RETURN_TEXT_P(cstring_to_text("OK"));
}

PG_FUNCTION_INFO_V1(mylink_disconnect);
Datum
mylink_disconnect(PG_FUNCTION_ARGS)
{
    RemoteConn *tconn;
    char *conn_name;

    if (PG_NARGS() == 1) {
        conn_name = text_to_cstring(PG_GETARG_TEXT_PP(0));
    } else {
        conn_name = default_conn_name;
    }

    mylink_dummy_func();

    tconn = getConnectionByName(conn_name);
    if (tconn) {
        deleteConnection(conn_name);
        mysql_close(&tconn->conn);
        pfree(tconn);
     }

    PG_RETURN_TEXT_P(cstring_to_text("OK"));
}

// SELECT COUNT(*) FROM aaa WHERE bb='cc' AND dd='ee'
/* PG_FUNCTION_INFO_V1(mylink_count_row); */
/* Datum */
/* mylink_count_row(PG_FUNCTION_ARGS) */
/* { */
/*     char *count_sql = NULL; */
/*     int retval = 0; */
/*     int my_ret = 0; */
/*     MYSQL_RES *my_res; */
/*     MYSQL_ROW my_row; */
/*     char *end_ptr; */
  
/*     count_sql = text_to_cstring(PG_GETARG_TEXT_PP(0)); */

/*     if (pconn == 0) { */
/*         ereport(WARNING, (errcode(-1), errmsg("EMYLINK: mysql not connected."))); */
/*         return (Datum)-1; */
/*     } */

/*     my_ret = mysql_real_query(&pconn->conn, count_sql, strlen(count_sql)); */
/*     if (my_ret == 0) { */
/*         my_res = mysql_store_result(&pconn->conn); */
/*         if (my_res == NULL) { */
/*             ereport(WARNING, (errcode(-1), errmsg(mylink_last_error(conn_name)))); */
/*             return (Datum)-3; */
/*         } else { */
/*             if (mysql_num_rows(my_res) > 0) { */
/*                 my_row = mysql_fetch_row(my_res); */
/*                 end_ptr = NULL; */
/*                 if (my_row[0] == NULL) { */
/*                     retval = -7; */
/*                 } else { */
/*                     retval = strtol(my_row[0], &end_ptr, 10); */
/*                 } */
/*                 mysql_free_result(my_res); */
/*             } else { */
/*                 return (Datum)-4; */
/*             } */
/*         } */
/*     } else { */
/*         return (Datum)-2; */
/*     } */
  
/*     PG_RETURN_INT32(retval); */
/*     // return 0; */
/* } */

PG_FUNCTION_INFO_V1(mylink_query);
Datum
mylink_query(PG_FUNCTION_ARGS)
{
    RemoteConn *tconn;
    char *conn_name;

    FuncCallContext     *funcctx;
    int                  call_cntr;
    int                  max_calls;
    // TupleDesc            tupdesc;
    AttInMetadata       *attinmeta;
    TupleDesc            mytupdesc;
    AttInMetadata        *myattinmeta;

    MYSQL_RES *my_res = NULL;
    MYSQL_ROW my_row;
    MYSQL_FIELD *my_field;
    unsigned long *my_lengths;
    char *my_sql;
    int my_ret;
    int num_fields;
    int num_rows;
    int ci;
    int my_fid;
    char *my_fname;
    int pg_fid;
    char *emsg;

    if (PG_NARGS() == 2) {
        conn_name = text_to_cstring(PG_GETARG_TEXT_PP(1));
    } else {
        conn_name = default_conn_name;
    }

    tconn = getConnectionByName(conn_name);

    /* stuff done only on the first call of the function */
    if (SRF_IS_FIRSTCALL()) {
        MemoryContext   oldcontext;

        /* create a function context for cross-call persistence */
        funcctx = SRF_FIRSTCALL_INIT();

        /* switch to memory context appropriate for multiple function calls */
        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

        ///////////
        if (tconn == NULL) {
            MemoryContextSwitchTo(oldcontext);
            ereport(WARNING, (errcode(-1),errmsg("mysql not connected")));
            SRF_RETURN_DONE(funcctx);
        }

        // my_sql = "SELECT uid,regip FROM pre_common_member_status";
        my_sql = text_to_cstring(PG_GETARG_TEXT_PP(0));
        my_ret = mysql_real_query(&tconn->conn, my_sql, strlen(my_sql));
        if (my_ret != 0) {
            MemoryContextSwitchTo(oldcontext);
            emsg = mylink_last_error(conn_name);
            ereport(WARNING, (errcode(-1),errmsg(emsg)));
            SRF_RETURN_DONE(funcctx);
        }

        my_res = mysql_store_result(&tconn->conn);
        if (my_res == NULL) {
            MemoryContextSwitchTo(oldcontext);
            // emsg = mysql_error(&tconn->conn);
            emsg = mylink_last_error(conn_name);
            ereport(WARNING, (errcode(-1),errmsg(emsg)));
            SRF_RETURN_DONE(funcctx);
        }
      
        funcctx->user_fctx = my_res;
        num_rows = mysql_num_rows(my_res);
        num_fields = mysql_num_fields(my_res);
        funcctx->max_calls = num_rows;

        pg_fid = TEXTOID;
        mytupdesc = CreateTemplateTupleDesc(num_fields, false);
        // automatically build fields desc from MYSQL_RES
        for (ci = 0; ci < num_fields; ci++) {
            my_field = mysql_fetch_field(my_res);
            my_fname = my_field->name;
            my_fid = my_field->type;
	
            switch (my_fid) {
            case MYSQL_TYPE_DECIMAL:
                pg_fid = FLOAT8OID;
                break;
            case MYSQL_TYPE_TINY:
                pg_fid = INT2OID;
                break;
            case MYSQL_TYPE_SHORT:
                pg_fid = INT4OID;
                break;
            case MYSQL_TYPE_LONG:
                pg_fid = INT8OID;
                break;
            case MYSQL_TYPE_FLOAT:
                pg_fid = FLOAT8OID;
                break;
            case MYSQL_TYPE_DOUBLE:
                pg_fid = FLOAT8OID;
                break;
            case MYSQL_TYPE_NULL:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_TIMESTAMP:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_LONGLONG:
                pg_fid = INT8OID;
                break;
            case MYSQL_TYPE_INT24:
                pg_fid = INT4OID;
                break;
            case MYSQL_TYPE_DATE:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_TIME:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_DATETIME:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_YEAR:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_NEWDATE:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_VARCHAR:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_BIT:
                pg_fid = BITOID;
                break;
            case MYSQL_TYPE_NEWDECIMAL:
                pg_fid = FLOAT8OID;
                break;
            case MYSQL_TYPE_ENUM:
                pg_fid = FLOAT8OID;
                break;
            case MYSQL_TYPE_SET:
                pg_fid = ANYENUMOID;
                break;
            case MYSQL_TYPE_TINY_BLOB:
                break;
            case MYSQL_TYPE_MEDIUM_BLOB:
                break;
            case MYSQL_TYPE_LONG_BLOB:
                break;
            case MYSQL_TYPE_BLOB:
                break;
            case MYSQL_TYPE_VAR_STRING:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_STRING:
                pg_fid = TEXTOID;
                break;
            case MYSQL_TYPE_GEOMETRY:
                pg_fid = TEXTOID;
                break;
            default:
                pg_fid = TEXTOID;
                break;
            }

            TupleDescInitEntry(mytupdesc, (AttrNumber)(ci+1), my_fname, pg_fid, -1, 0);
        }
        myattinmeta = TupleDescGetAttInMetadata(mytupdesc);

        funcctx->attinmeta = myattinmeta;

        /* total number of tuples to be returned */
        // funcctx->max_calls = PG_GETARG_UINT32(0);

        /* Build a tuple descriptor for our result type */
        // if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
        /* ereport(ERROR, */
        /* 	(errcode(ERRCODE_FEATURE_NOT_SUPPORTED), */
        /* 	 errmsg("function returning record called in context " */
        /* 		"that cannot accept type record"))); */

        /* tupdesc = CreateTemplateTupleDesc(3, false); */
        /* TupleDescInitEntry(tupdesc, (AttrNumber) 1, "ff1", */
        /* 			 INT4OID, -1, 0); */
        /* TupleDescInitEntry(tupdesc, (AttrNumber) 2, "ff2", */
        /* 			 INT4OID, -1, 0); */
        /* TupleDescInitEntry(tupdesc, (AttrNumber) 3, "f23", */
        /* 			 INT4OID, -1, 0); */
      
        /*
         * generate attribute metadata needed later to produce tuples from raw
         * C strings
         */
        // attinmeta = TupleDescGetAttInMetadata(tupdesc);
        // funcctx->attinmeta = attinmeta;

        MemoryContextSwitchTo(oldcontext);
    }

    /* stuff done on every call of the function */
    funcctx = SRF_PERCALL_SETUP();

    call_cntr = funcctx->call_cntr;
    max_calls = funcctx->max_calls;
    attinmeta = funcctx->attinmeta;

    my_res = funcctx->user_fctx;
    num_rows = mysql_num_rows(my_res);
    num_fields = mysql_num_fields(my_res);

    if (call_cntr < max_calls) {    /* do when there is more left to send */
        /* char       **values; */
        /* HeapTuple    tuple; */
        /* Datum        result; */

        char ** my_values;
        HeapTuple  mytuples;
        Datum    myresult;

        my_row = mysql_fetch_row(my_res);
        if (my_row == NULL) {
            mysql_free_result(my_res);
            SRF_RETURN_DONE(funcctx);	
        }

        my_lengths = mysql_fetch_lengths(my_res);
        my_values = (char**)palloc(num_fields * sizeof(char*));
        for (ci = 0; ci < num_fields; ci++) {
            if (my_row[ci] == NULL) {
                my_values[ci] = NULL;
            } else {
                my_values[ci] = (char*)palloc(my_lengths[ci] * sizeof(char)+1);
                strncpy(my_values[ci], my_row[ci], my_lengths[ci]*sizeof(char)+1);
            }
        }

        mytuples = BuildTupleFromCStrings(attinmeta, my_values);
        myresult = HeapTupleGetDatum(mytuples);

        // cleanup
        for (ci = 0; ci < num_fields; ci++) {
            if (my_values[ci] != NULL) {
                pfree(my_values[ci]);
            }
        }
        pfree(my_values);

        SRF_RETURN_NEXT(funcctx, myresult);

        /*
         * Prepare a values array for building the returned tuple.
         * This should be an array of C strings which will
         * be processed later by the type input functions.
         */
        /* values = (char **) palloc(3 * sizeof(char *)); */
        /* values[0] = (char *) palloc(16 * sizeof(char)); */
        /* values[1] = (char *) palloc(16 * sizeof(char)); */
        /* values[2] = (char *) palloc(16 * sizeof(char)); */

        /* snprintf(values[0], 16, "%d", 1 * PG_GETARG_INT32(1)); */
        /* snprintf(values[1], 16, "%d", 2 * PG_GETARG_INT32(1)); */
        /* snprintf(values[2], 16, "%d", 3 * PG_GETARG_INT32(1)); */

        /* build a tuple */
        // tuple = BuildTupleFromCStrings(attinmeta, values);

        /* make the tuple into a datum */
        // result = HeapTupleGetDatum(tuple);

        /* clean up (this is not really necessary) */
        /* pfree(values[0]); */
        /* pfree(values[1]); */
        /* pfree(values[2]); */
        /* pfree(values); */

        /* SRF_RETURN_NEXT(funcctx, result); */
    } else {  /* do when there is no more left */
        mysql_free_result(my_res);
        SRF_RETURN_DONE(funcctx);
    }
}

PG_FUNCTION_INFO_V1(mylink_exec);
Datum
mylink_exec(PG_FUNCTION_ARGS)
{
    RemoteConn *tconn;
    char *conn_name;
    char *my_sql;
    int  my_ret;

    if (PG_NARGS() == 2) {
        conn_name = text_to_cstring(PG_GETARG_TEXT_PP(1));
    } else {
        conn_name = default_conn_name;
    }

    my_sql = text_to_cstring(PG_GETARG_TEXT_PP(0));

    tconn = getConnectionByName(conn_name);
    if (tconn) {
        my_ret = mysql_real_query(&tconn->conn, my_sql, strlen(my_sql));
        if (my_ret != 0) {
            ereport(WARNING, (errcode(-1), errmsg(mylink_last_error(conn_name))));
            PG_RETURN_TEXT_P(cstring_to_text("ERR"));
        }
    } else {
    }
    
    PG_RETURN_TEXT_P(cstring_to_text("OK"));
}


PG_FUNCTION_INFO_V1(mylink_select_db);
Datum
mylink_select_db(PG_FUNCTION_ARGS)
{
    RemoteConn *tconn;
    int my_ret;
    char *dbname;
    char *conn_name;

    if (PG_NARGS() == 2) {
        conn_name = text_to_cstring(PG_GETARG_TEXT_PP(1));
    } else {
        conn_name = default_conn_name;
    }

    dbname = text_to_cstring(PG_GETARG_TEXT_PP(0));
    if (dbname == NULL) {
        PG_RETURN_TEXT_P(cstring_to_text("ERR"));
    }

    tconn = getConnectionByName(conn_name);

    my_ret = mysql_select_db(&tconn->conn, dbname);
    if (my_ret != 0) {
        ereport(WARNING, (errcode(-1), errmsg(mylink_last_error(conn_name))));
        PG_RETURN_TEXT_P(cstring_to_text("ERR"));
    }

    PG_RETURN_TEXT_P(cstring_to_text("OK"));
}

/*
 * List all open dblink connections by name.
 * Returns an array of all connection names.
 * Takes no params
 */
PG_FUNCTION_INFO_V1(mylink_get_connections);
Datum
mylink_get_connections(PG_FUNCTION_ARGS)
{
	HASH_SEQ_STATUS status;
	RemoteConnHashEnt *hentry;
	ArrayBuildState *astate = NULL;

	if (hconn) {
        hash_seq_init(&status, hconn);
        while ((hentry = (RemoteConnHashEnt *) hash_seq_search(&status)) != NULL) {
            /* stash away current value */
            if (strcmp(hentry->name, default_conn_name) == 0) {
                astate = accumArrayResult(astate,
                                          cstring_to_text(""),
                                          false, TEXTOID, CurrentMemoryContext);
            } else {
                astate = accumArrayResult(astate,
                                          CStringGetTextDatum(hentry->name),
                                          false, TEXTOID, CurrentMemoryContext);
            }
        }
    }

	if (astate)
		PG_RETURN_ARRAYTYPE_P(makeArrayResult(astate,
											  CurrentMemoryContext));
	else
		PG_RETURN_NULL();
}


///////////internal
static RemoteConn * getConnectionByName(const char *name)
{
	RemoteConnHashEnt *hentry;
	char	   *key;

	if (!hconn)
		hconn = createConnHash();

	key = pstrdup(name);
	truncate_identifier(key, strlen(key), false);
	hentry = (RemoteConnHashEnt *) hash_search(hconn,
											   key, HASH_FIND, NULL);

	if (hentry)
		return (hentry->rconn);

	return (NULL);
}

static HTAB *createConnHash(void)
{
	HASHCTL		ctl;

	ctl.keysize = NAMEDATALEN;
	ctl.entrysize = sizeof(RemoteConnHashEnt);

	return hash_create("mylink hash", NUMCONN, &ctl, HASH_ELEM);
}

static void createNewConnection(const char *name, RemoteConn *rconn)
{
	RemoteConnHashEnt *hentry;
	bool		found;
	char	   *key;

	if (!hconn)
		hconn = createConnHash();

	key = pstrdup(name);
	truncate_identifier(key, strlen(key), true);
	hentry = (RemoteConnHashEnt *) hash_search(hconn, key,
											   HASH_ENTER, &found);

	if (found) {
        // PQfinish(rconn->conn);
        mysql_close(&rconn->conn);
        pfree(rconn);

        ereport(ERROR,
                (errcode(ERRCODE_DUPLICATE_OBJECT),
                 errmsg("duplicate connection name")));
    }

	hentry->rconn = rconn;
	strlcpy(hentry->name, name, sizeof(hentry->name));
}

static void deleteConnection(const char *name)
{
	RemoteConnHashEnt *hentry;
	bool		found;
	char	   *key;

	if (!hconn)
		hconn = createConnHash();

	key = pstrdup(name);
	truncate_identifier(key, strlen(key), false);
	hentry = (RemoteConnHashEnt *) hash_search(hconn,
											   key, HASH_REMOVE, &found);

	if (!hentry)
		ereport(ERROR,
				(errcode(ERRCODE_UNDEFINED_OBJECT),
				 errmsg("undefined connection name")));

}


